#include <algorithm>
#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <ctime>

#include "TH1.h"
#include "TF1.h"
#include "TRandom3.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TFile.h"

#define numScrMaps 50000

using namespace std;
//g++ analysis_main.cpp `root-config --cflags` `root-config --glibs` -o analysisMain

//sort array in increasing values
int partition(double tablica[], int p, int r) // divide the tables into two parts, in the first all numbers are smaller or equal x, in the second bigger or equal to x
{
    double x = tablica[p],w; // we choose x
    int i = p, j = r; // i, j - indexes in the array
    while (true) // infinite loop - leave it only by return j
        {
        while (tablica[j] > x)
            j--;
        while (tablica[i] < x)
            i++;
        if (i < j) // we change places when i <j
        {
            w = tablica[i];
            tablica[i] = tablica[j];
            tablica[j] = w;
            i++;
            j--;
        }
        else // when i> = j we return j as the partition point of the array
            return j;
        }
}

void quicksort(double tablica[], int p, int r) // qsort
{
    int q;
    if (p < r)
    {
        q = partition(tablica,p,r); // we divide the arrays into two parts; q is the partition point
        quicksort(tablica, p, q); // we recursively call the quicksort for the first part of the array
        quicksort(tablica, q+1, r); // we recursively call the quicksort for the second part of the array
    }
}
////////////////////////////////////////////////////

//calculate how many times two events are found within a 5 minutes time window
void Doublet_funct(const int numOfEvents, double timeStamp[], double counterNumOfDoublets_SC[], int i, int timeWindow)
{
    int j = 0;
    while (j <= numOfEvents - 2)
    {
	    if (timeStamp[j+1] - timeStamp[j] <= timeWindow)
        {
            counterNumOfDoublets_SC[i]++;
        }

        j++;
    }
}

//finds the smallest value in an array
double min_array(double array[], int numOfScrMaps)
{
    double min = array[0];

    for (int i = 0; i < numOfScrMaps; i++)
    {
        if (array[i] < min && array[i] != 0)
        {
            min = array[i];
        }
    }

    return min;
}

//finds the greatest value in an array
double max_array(double array[], int numOfScrMaps)
{
    double max = array[0];

    for (int i = 0; i < numOfScrMaps; i++)
    {
        if (array[i] > max)
        {
            max = array[i];
        }
    }

    return max;
}

/*obtain parameters of doublet distribution of scrambled maps and calculate significance of number
of doublets found in data*/
void fit_histo_doublet(TH1D* histo, double counterNumOfDoublets_data, double parameters_doublet[])
{
    double mean = histo->GetMean();
    double std = histo->GetRMS();

    parameters_doublet[0] = (counterNumOfDoublets_data - mean)/std;
    parameters_doublet[1] = mean + 3*std;
    parameters_doublet[2] = mean + 5*std;
}

int main(int nargv, char **argv)
{
    time_t start,end;
    time (&start);
	// extracting arguments
	int timeWindow = atoi(argv[1]);
    string user_name = argv[2];
    int user_id = atoi(argv[3]);
    int day = atoi(argv[4]);

    // Input file - timestamps
	ifstream input;
  	char data [256];
  	sprintf(data,"timestamp_%d_%d.txt",user_id,day); //user id is defined in the bash script

  	cout << "Processing data from user " << user_name.c_str() << " - period " << day << "..." << endl;

  	// Output file containing all the parameters needed for the plots (test statistic, number of doublets, sigma values, etc...)
  	FILE *pfile;
  	char output[256];
 	sprintf(output,"output_parameters_%s.txt",user_name.c_str()); //output file
 	pfile = fopen(output,"a");

 	int numOfEventsInFile = 0;
  	double time_of_evt;
    vector<double> timeStamp_Data_Vec; //vector storing the timestamps

 	input.open(data);
  	//Checking if the data file exists or/and can be opened
 	if (!input)
 	{
   		cout << "Unable to open data file." << endl;
   		return 0;
 	}

 	while (input >> time_of_evt)
 	{
        timeStamp_Data_Vec.push_back(time_of_evt);
   		numOfEventsInFile++; // Counting the number of events in data file
 	}
 	input.close();

 	cout << "Number of events in data: " << numOfEventsInFile << endl;

 	//removing all the events with the same timestamp for the test statistic analysis
 	for (int i = 1; i < timeStamp_Data_Vec.size(); i++)
 	{
 		if (timeStamp_Data_Vec[i] == timeStamp_Data_Vec[i-1])
 		{
 			timeStamp_Data_Vec.erase(timeStamp_Data_Vec.begin() + i);
 			i--;
 		}
 	}

    const int numOfEvents = timeStamp_Data_Vec.size();

    cout << "Number of events in data after removing events with same timestamps: " << numOfEvents << endl;

 	// storing the timestamps in an array and converting them to seconds
    double timeStamp_Data[numOfEvents];
    for(int i = 0; i < numOfEvents; i++)
    {
        timeStamp_Data[i] = timeStamp_Data_Vec[i]/1000;
    }

    double firstTime = timeStamp_Data[0], lastTime = timeStamp_Data[numOfEvents-1]; // saving first and last timestamp of data
    double timeSpan = lastTime - firstTime; // time duration between first and last detection
    
    double ontime = 86400;

    cout << setprecision(10) << "Time covered by data = " << timeSpan << " sec" << endl;
    cout << setprecision(10) << "Real ontime = " << ontime << " sec" << endl;

	// Scrambled maps (SM) variables
	TRandom *r1 = new TRandom3();
	double rand1;
	double timeStamp_SC[numOfEvents]; // 2 dimensional array storing randomly generated timestamps for each SM
	double counterNumOfDoublets_SC[numScrMaps]; //array storing the number of doublets found in each SM

	//*********** Generating the scrambled maps in a 2-D array (numScrMaps by numOfEvents)**********
	for (int i = 0; i < numScrMaps; i++)
	{
		counterNumOfDoublets_SC[i] = 0;

		for (int j = 0; j < numOfEvents; j++)
		{
			rand1 = r1->Rndm(j);
            timeStamp_SC[j] = r1->Uniform(0,ontime);
		}

		// sorting arrays rows in increasing timestamps
        quicksort(timeStamp_SC,0,numOfEvents-1);

		// performing doublet analysis
        Doublet_funct(numOfEvents,timeStamp_SC,counterNumOfDoublets_SC,i,timeWindow);
	}
	//**********************************************************************************************

	double maxBin_DOUBLET = max_array(counterNumOfDoublets_SC,numScrMaps); // variable to store the highest number of doublets for histogram purpose
	double minBin_DOUBLET = min_array(counterNumOfDoublets_SC,numScrMaps); // variable to store the lowest number of doublets for histogram purpose (arbitrary high enough value given)

	//******************doublet histograms**************************************

	TCanvas *c1 = new TCanvas("c1","histogram");

	// Doublet
    TString histo_file = TString::Format("SM_doublet_distribution_user-%d_%d.root",user_id,day);
    TFile file1(histo_file, "RECREATE");

	TString histoDOUBLET = TString::Format("histogram_doublets");
	TString histoDOUBLET_title = TString::Format("Distribution of number of doublets in background - User %s",user_name.c_str());
	TH1D *h1 = new TH1D(histoDOUBLET, histoDOUBLET_title, 100, minBin_DOUBLET-1, maxBin_DOUBLET+1);

 	for (int i = 0; i < numScrMaps; i++)
	{
		h1->Fill(counterNumOfDoublets_SC[i]);
	}

	h1->GetXaxis()->SetTitle("Number of doublets");

	h1->Write();
	//**********************************************************************************************

	//***********Unblinding data: calculating number of doublets*****************
	double counterNumOfDoublets_Data_array[1];

	counterNumOfDoublets_Data_array[0] = 0;

	Doublet_funct(numOfEvents,timeStamp_Data,counterNumOfDoublets_Data_array,0,timeWindow);
	//**********************************************************************************************

	double counterNumOfDoublets_Data = counterNumOfDoublets_Data_array[0];

	//*******************Getting 3 and 5 sigma values for TS and Doublet histograms*****************
	double parameters_TS[3];
	double parameters_doublet[3];

	// fitting histograms to obtain sigma and p values

	fit_histo_doublet(h1,counterNumOfDoublets_Data,parameters_doublet);

	double sig_doublet = parameters_doublet[0], threeSig_doublet = parameters_doublet[1], fiveSig_doublet = parameters_doublet[2];
	double pvalue_doublet = TMath::Erf(sig_doublet/sqrt(2));
	double meanNumOfDoublets_SC = h1->GetMean();

	cout << "----- Doublet analysis -----" << endl;
	cout << "Expected number of doublets = " << meanNumOfDoublets_SC << " || Number of doublets in data = " << counterNumOfDoublets_Data << " || pvalue = " << pvalue_doublet << " || sigma = " << sig_doublet << endl;
	cout << "3 sigma at " << threeSig_doublet << " || 5 sigma at " << fiveSig_doublet << endl;

    double excess = counterNumOfDoublets_Data - meanNumOfDoublets_SC;

	// saving all parameters needed for plots in output file
    fprintf(pfile, "%f\t%f\t%f\t%f\t%f\t%f\n",meanNumOfDoublets_SC,counterNumOfDoublets_Data,sig_doublet,threeSig_doublet,fiveSig_doublet,excess);
	time (&end);
    double dif = difftime (end,start);
    printf ("Elasped time is %.2lf seconds.\n\n", dif );
}